searchData={"content_type":"text/markdown","items":[{"doc":"# Cinema\n\nCinema is a simple Elixir framework for managing incremental materialized views entirely in Elixir/Ecto.","ref":"Cinema.html","title":"Cinema","type":"module"},{"doc":"Cinema can be installed by adding `cinema` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:cinema, \"~> 0.1.0\"}\n  ]\nend\n```\n\nCinema has an optional dependency on [Oban Pro](https://getoban.pro) as an alternate runtime for materializing projection graphs. Oban Pro support is automatically enabled if `Hex` detects the `oban` repo in your global setup.\n\nPlease see the [Oban Pro](https://getoban.pro) documentation for more information on how to install and configure Oban Pro.","ref":"Cinema.html#module-installation","title":"Installation - Cinema","type":"module"},{"doc":"Cinema introduces two basic concepts:\n\n- **Projections**: A projection is a behaviour that allows you to declaratively define instructions in Elixir for how to derive rows to write into your materialized views. They statically define all `c:inputs/0` (other projections, if needed), `c:output/1` (usually an `Ecto.Query` or stream passed into subsequent projections), and a `c:derivation/2` callback which is run in tandem with all inputs to produce output.\n- **Lenses**: A lens is a struct which contains filters and other options which can be used to modify the \"scope\" of what a projection is required to derive. In simple use cases, you can think of lenses as automatically applying filters such as: `where: x.org_id == ^org_id` to the outputs of all input projections automatically.\n\nWhen you want to actually incrementally rematerialize a view, you create a `Cinema.Lens.t()` (or a simple keyword list for simple filters), and pass that into the `Cinema.project/3` function like so:\n\n```elixir\niex> Cinema.project(MyApp.Projections.AccountsReceivable, [org_id: 123, date: ~D[2022-01-01]])\n[\n  %MyApp.Projections.AccountsReceivable{\n    org_id: 123,\n    date: ~D[2022-01-01],\n    ...\n  },\n  ...\n]\n```\n\nProjections generally define their own `Ecto.Schema` internally and can also be queried directly -- note that this will not rematerialize any dependencies or rows in the table you're querying:\n\n```elixir\niex> MyApp.Repo.all(MyApp.Projections.AccountsReceivable)\n[\n  %MyApp.Projections.AccountsReceivable{\n    org_id: 123,\n    date: ~D[2022-01-01],\n    ...\n  },\n  ...\n]\n```\n\nProjections can include other projections as inputs, and Cinema will automatically rematerialize those projections as needed. For example, if `AccountsReceivable` depends on `Invoices`, Cinema will automatically rematerialize `Invoices` before rematerializing `AccountsReceivable`.\n\nProjection graphs usually begin with \"virtual\" projections that have no inputs or `derivation/2` callback, instead only outputting either an `Ecto.Query` or stream which is passed directly through any `Cinema.Lens.t()` and into the next projection in the graph.\n\nCinema does this by building a DAG of all projections and their dependencies. Cinema will likewise try to run any projections in parallel where possible. A minimal example of a projection looks like the following:\n\n```elixir\ndefmodule MyApp.Projections.Accounts do\n  use Cinema.Projection, virtual?: true\n\n  @impl Cinema.Projection\n  def inputs, do: []\n\n  @impl Cinema.Projection\n  def output, do: from(a in \"accounts\", select: a.id)\nend\n\ndefmodule MyApp.Projections.AccountsReceivable do\n  use Cinema.Projection,\n    conflict_target: [:account_id],\n    required_fields: [:account_id],\n    on_conflict: :replace_all,\n    read_repo: MyApp.Repo.Replica,\n    write_repo: MyApp.Repo,\n    timeout: :timer.minutes(5),\n\n  alias Cinema.Projection\n  alias MyApp.Projections.Accounts\n\n  @primary_key false\n  schema \"accounts_receivable\" do\n    field(account_id:, :id)\n    field(total, :integer)\n\n    timestamps()\n  end\n\n  @impl Cinema.Projection\n  def inputs, do: [Accounts]\n\n  @impl Cinema.Projection\n  def output, do: from(a in \"accounts_receivable\", select: a)\n\n  @impl Cinema.Projection\n  def derivation({Accounts, stream}, lens) do\n    Projection.dematerialize(lens)\n\n    stream\n    |> Stream.chunk_every(2000)\n    |> Stream.map(&from x in MyApp.Invoice, where: x.account_id in ^&1, select: %{account_id: x.account_id, total: sum(x.total)})\n    |> Stream.map(&Projection.materialize/1)\n    |> Stream.run()\n  end\nend\n```","ref":"Cinema.html#module-usage","title":"Usage - Cinema","type":"module"},{"doc":"Currently, Cinema lets you configure the following options:\n\n- `:engine` - The runtime to use for executing projection graphs. Defaults to `Cinema.Engine.Task`.\n- `:async` - Whether to run projections asynchronously. Defaults to `true`.\n\nAdditional configuration options can be implemented on a projection-by-projection basis, please see the docs for the `Cinema.Projection` behaviour for more information.","ref":"Cinema.html#module-configuration","title":"Configuration - Cinema","type":"module"},{"doc":"Cinema is released under the [MIT License](LICENSE.md).","ref":"Cinema.html#module-license","title":"License - Cinema","type":"module"},{"doc":"","ref":"Cinema.html#fetch/1","title":"Cinema.fetch/1","type":"function"},{"doc":"Projects a module that implements the `Cinema.Projection` behaviour.\n\nProjection is done by reflecting on the given projection and building a dependency graph of all of it's inputs,\nwhich is then executed in the correct order.\n\nThe output of each dependency projection is streamed into the given projection's `c:Cinema.Projection.derivation/2`\ncallback. This callback is responsible for materializing database table rows which can then be returned as the final\noutput of the projection.\n\nOptionally takes a `Lens.t()` or a `Keyword.t()` filter list to apply to the projection.\n\nAdditionally, you can pass in a `Keyword.t()` list of options to control the behavior of how the projection is\nexecuted.\n\nNote that by default, the given projection's final `c:Cinema.Projection.output/0` will be awaited and\nevaluated by the projection's configured `read_repo`. If you want to control the execute the projection, pass\n`async: true` as an option.","ref":"Cinema.html#project/3","title":"Cinema.project/3","type":"function"},{"doc":"* `:async` - Defaults to `false`. If `true`, the projection will be executed immediately but the given projection's\n  final output will not be awaited and returned. Instead, the projection itself will be returned.\n\n  You can then await and fetch the final output via `Cinema.fetch/1`.\n\n* `:engine` - The engine to use to execute the projection. Defaults to `Cinema.Engine.Task`. See\n  `Cinema.Engine` for more information.\n\n* `:timeout` - The timeout to use when executing the projection. Defaults to `1 minute`.\n\n* `:allow_empty_filters` - Defaults to `false`. If `true`, skips the warning message that gets logged when an empty\n  filter list is provided.\n\nOptions are additionally also passed to the engine that is used to execute the projection, as well as used when\nbuilding the projection itself.","ref":"Cinema.html#project/3-options","title":"Options - Cinema.project/3","type":"function"},{"doc":"The `Cinema.Engine` module provides a behaviour for defining engines which can execute projections\ndefined using the `Cinema.Projection` module.\n\nSee the `Cinema.Engine.Task` module for an example implementation of this behaviour.\nAlso see the `Cinema.Projection` module for more details re: defining projections.","ref":"Cinema.Engine.html","title":"Cinema.Engine","type":"behaviour"},{"doc":"","ref":"Cinema.Engine.html#c:exec/1","title":"Cinema.Engine.exec/1","type":"callback"},{"doc":"Executes the given `Projection.t()` using the given engine. Returns `{:ok, projection}` if the projection\nwas successfully executed, otherwise returns `{:error, projection}`.\n\nYou can fetch the output of the `Projection.t()` (assuming it was successfully executed) by passing the\nreturned `Projection.t()` to `Cinema.Projection.fetch/1` function.","ref":"Cinema.Engine.html#exec/2","title":"Cinema.Engine.exec/2","type":"function"},{"doc":"Fetches the output of a `Projection.t()` which has been executed using `Cinema.Engine.exec/2`.","ref":"Cinema.Engine.html#fetch/1","title":"Cinema.Engine.fetch/1","type":"function"},{"doc":"","ref":"Cinema.Engine.html#c:fetch/1","title":"Cinema.Engine.fetch/1","type":"callback"},{"doc":"Returns `true` if the given module implements the `Elixir.Cinema.Engine` behaviour, otherwise returns `false`.\n\nSee examples:\n\n  ```elixir\n  iex> Cinema.Engine.implemented?(Enum)\n  false\n  iex> Cinema.Projection.implemented?(Cinema.Engine.Task)\n  true\n  ```","ref":"Cinema.Engine.html#implemented?/1","title":"Cinema.Engine.implemented?/1","type":"function"},{"doc":"","ref":"Cinema.Engine.html#test_mode/1","title":"Cinema.Engine.test_mode/1","type":"function"},{"doc":"","ref":"Cinema.Engine.html#test_mode?/0","title":"Cinema.Engine.test_mode?/0","type":"function"},{"doc":"The `Cinema.Engine.Task` module provides a simple implementation of the `Cinema.Engine` behaviour\nwhich uses the `Task` module to execute projections.\n\nWhen in test mode, all Task functions will be executed synchronously. You can enable test mode by calling\n`Cinema.Utils.Task.test_mode(true)` somewhere before executing projections.\n\nSee the `Cinema.Engine` module for more details re: defining engines.","ref":"Cinema.Engine.Task.html","title":"Cinema.Engine.Task","type":"module"},{"doc":"Executes the given `Projection.t()` using the Task engine. Returns `{:ok, projection}` if the projection\nwas successfully executed, otherwise returns `{:error, projection}`.\n\nYou can fetch the output of the `Projection.t()` (assuming it was successfully executed) by passing the\nreturned `Projection.t()` to `Cinema.Projection.fetch/1` function.","ref":"Cinema.Engine.Task.html#exec/2","title":"Cinema.Engine.Task.exec/2","type":"function"},{"doc":"* `:timeout` - The timeout to use when executing the projection. Defaults to `1 minute`.\n* `:skip_dependencies` - Defaults to `false`. If `true`, skips executing the dependencies of the given projection.\n  this is very useful for testing purposes where you use `ExMachina` to \"mock\" the results of all input projections.","ref":"Cinema.Engine.Task.html#exec/2-options","title":"Options - Cinema.Engine.Task.exec/2","type":"function"},{"doc":"Fetches the output of a `Projection.t()` which has been executed using `Cinema.Engine.exec/2`.","ref":"Cinema.Engine.Task.html#fetch/1","title":"Cinema.Engine.Task.fetch/1","type":"function"},{"doc":"This module is responsible to defining \"projections\", which are similar to \"views\" built into\nmost RDBMS systems.","ref":"Cinema.Projection.html","title":"Cinema.Projection","type":"behaviour"},{"doc":"A projection is simply a module which tells `Cinema` three things:\n\n1) What inputs (usually other projections) are required to execute a projection.\n2) What outputs (usually a `Stream`-compatible term) are produced by executing a projection.\n3) Most importantly, the instructions required to derive data for the projection.\n\nIt is important to call out that a `Projection` is different from defining a `View` in a traditional\nRDBMS system. A `View` is a stored query that is executed on-demand, while a `Projection` is a\nseries of instructions that can be executed to derive a result, which is optionally stored in a table\nfor querying.\n\nWhile `View`s can be materialized, at least in PostgreSQL, you're usually forced to refresh the\nentire view whenever you want to update any stored data. `Projection`s on the other hand are designed\nto allow some form of incremental materialization, where only data relevent to the `Cinema.Lens`\nis required to be refreshed.\n\nCurrently, Projections need to be manually called to be refreshed; though future work may include\nautomatic triggers for refreshing Projections when their inputs change.\n\nPlease see `Cinema.project/2` for more details on how projections are executed.","ref":"Cinema.Projection.html#module-key-concepts","title":"Key Concepts - Cinema.Projection","type":"behaviour"},{"doc":"Projections currently come in two flavours: virtual and materialized, defaulting to materialized\nfor most cases.\n\nA materialized projection requires users to define an `Ecto.Schema` schema and associated database\nmigration to store the projection's output. These projections can then be indexed and queried like any\nother schema.\n\nVirtual projections on the other hand are not stored in the database, and are instead derived on-the-fly\nby executing the projection's instructions (if any) and immediately returning the Projection's output.\n\nGenerally, virtual projections are used for kickstarting larger `Cinema` data pipelines from\nyour application's existing database tables, while materialized projections are used for storing\nintermediate results for later querying.\n\nYou can control whether a projection is materialized or virtual by setting the `:materialized` option\nto `true` or `false` respectively. It is set to `true` by default.","ref":"Cinema.Projection.html#module-materialized-vs-virtual-projections","title":"Materialized vs Virtual Projections - Cinema.Projection","type":"behaviour"},{"doc":"Materialized Projections are automatically dematerialized prior to materialization, which means that\nthe projection's output is deleted before the projection is executed. This is useful for ensuring that\nthe projection's output is always up-to-date, and can be used to ensure that the projection's output\nis idempotent.\n\nOnly data which is in \"scope\" of a given Projection and Lens is dematerialized, so generally speaking,\nassuming `(dematerialize(input, lens); materialize(input, lens)) == materialize(input, lens)`, you\ndon't have to worry about dematerialization causing data loss.\n\nIf you have a projection that is not idempotent, you can disable dematerialization by setting the\n`:dematerialize` option to `false`, though you should be aware that you may want to manually manage\nthe projection's output to ensure that it is up-to-date and any outdated data is cleaned up.","ref":"Cinema.Projection.html#module-dematerialization","title":"Dematerialization - Cinema.Projection","type":"behaviour"},{"doc":"Projections can also, optionally, trigger side-effects by writing to the database or other external\nsystems. This can be done by executing the desires side-effect causing code within a Projection's\n`derive/2` callback.\n\nGenerally in such projections, it is inadvisable to dematerialize the projection, as the side-effects\nmay not be idempotent, and may rely on the state of the Projection's output as a log of what has\nalready been processed.\n\nIf this is the cases, dematerialization can be disabled by setting the `:dematerialize` option to `false`.","ref":"Cinema.Projection.html#module-side-effects","title":"Side Effects - Cinema.Projection","type":"behaviour"},{"doc":"All Projections emit Telemetry events when they are executed, which can be used to monitor said projections.\n\nThe `Cinema` Telemetry events are as follows: TBA","ref":"Cinema.Projection.html#module-telemetry","title":"Telemetry - Cinema.Projection","type":"behaviour"},{"doc":"Builds a new instance of the given projection, with the given lens. Does not execute the projection.\nSee `Cinema.Engine` for functions that operate on `Projection.t()`s.\n\nTakes an optional `lens` argument, which is a `Cinema.Lens.t()` struct that can be used to\nscope the projection's inputs and outputs.\n\nAlso takes an optional `opts` argument, which is a `Keyword.t()` list of options that can be used to\nconfigure the projection's behavior.","ref":"Cinema.Projection.html#build!/3","title":"Cinema.Projection.build!/3","type":"function"},{"doc":"* `:application` - The application to use when fetching the list of all projections. Defaults to `nil`.","ref":"Cinema.Projection.html#build!/3-options","title":"Options - Cinema.Projection.build!/3","type":"function"},{"doc":"","ref":"Cinema.Projection.html#dematerialize/1","title":"Cinema.Projection.dematerialize/1","type":"macro"},{"doc":"Returns the dependency graph needed to execute the given projection.","ref":"Cinema.Projection.html#dependencies/2","title":"Cinema.Projection.dependencies/2","type":"function"},{"doc":"","ref":"Cinema.Projection.html#c:derivation/2","title":"Cinema.Projection.derivation/2","type":"callback"},{"doc":"","ref":"Cinema.Projection.html#c:fields/0","title":"Cinema.Projection.fields/0","type":"callback"},{"doc":"Returns `true` if the given module implements the `Elixir.Cinema.Projection` behaviour, otherwise returns `false`.\n\nSee examples:\n\n  ```elixir\n  iex> Cinema.Projection.implemented?(Enum)\n  false\n\n  iex> defmodule MyApp.SomeProjection do\n  ...>   use Cinema.Projection\n  ...>\n  ...>   @impl Cinema.Projection\n  ...>   def inputs, do: []\n  ...>\n  ...>   @impl Cinema.Projection\n  ...>   def ouput, do: []\n  ...> end\n  iex> Cinema.Projection.implemented?(MyApp.SomeProjection)\n  true\n  ```","ref":"Cinema.Projection.html#implemented?/1","title":"Cinema.Projection.implemented?/1","type":"function"},{"doc":"","ref":"Cinema.Projection.html#c:inputs/0","title":"Cinema.Projection.inputs/0","type":"callback"},{"doc":"Lists all defined projections defined in the given application or loaded in the VM.","ref":"Cinema.Projection.html#list/1","title":"Cinema.Projection.list/1","type":"function"},{"doc":"","ref":"Cinema.Projection.html#materialize/1","title":"Cinema.Projection.materialize/1","type":"macro"},{"doc":"","ref":"Cinema.Projection.html#c:output/0","title":"Cinema.Projection.output/0","type":"callback"},{"doc":"Returns `true` if the given module is a virtual projection (does not define a schema), otherwise returns `false`.\n\nSee examples:\n\n  ```elixir\n  iex> Cinema.Projection.virtual?(Enum)\n  false\n\n  iex> defmodule MyApp.VirtualProjection do\n  ...>   use Cinema.Projection\n  ...>\n  ...>   @impl Cinema.Projection\n  ...>   def inputs, do: []\n  ...>\n  ...>   @impl Cinema.Projection\n  ...>   def ouput, do: []\n  ...> end\n  iex> Cinema.Projection.virtual?(MyApp.VirtualProjection)\n  true\n  ```","ref":"Cinema.Projection.html#virtual?/1","title":"Cinema.Projection.virtual?/1","type":"function"},{"doc":"","ref":"Cinema.Projection.html#t:implementation/0","title":"Cinema.Projection.implementation/0","type":"type"},{"doc":"","ref":"Cinema.Projection.html#t:input/0","title":"Cinema.Projection.input/0","type":"type"},{"doc":"","ref":"Cinema.Projection.html#t:output/0","title":"Cinema.Projection.output/0","type":"type"},{"doc":"","ref":"Cinema.Projection.html#t:t/0","title":"Cinema.Projection.t/0","type":"type"},{"doc":"The `Elixir.Cinema.Projection.Lens` module provides a struct for defining lenses which can be used to manipulate\nthe inputs to a given projection.\n\nLenses can be used to filter, scope, or otherwise manipulate the inputs to a given projection.\n\nSee the `Elixir.Cinema.Projection.Lens.apply/3` and `Elixir.Cinema.Projection.Lens.build!/2` functions for more information.","ref":"Cinema.Projection.Lens.html","title":"Cinema.Projection.Lens","type":"module"},{"doc":"Uses the given `Elixir.Cinema.Projection.Lens.t()` to manipulate inputs to the given projection.\n\nIf the given `Elixir.Cinema.Projection.Lens.t()` has a `reducer` function, a given input projection's output\nwill be reduced over with the given `filters`, allowing you to customize the behaviour of how\na projection's inputs are scoped or modified.\n\nIf the given `Elixir.Cinema.Projection.Lens.t()` does not have a `reducer` function, the given `filters` will\nbe used to filter the given input projection's output the same way `Ecto.Query.where/3` does\nwhen given a static `Keyword.t()` as a final argument.\n\nReturns either the manipulated queryable, or a stream capturing said queryable depending on if\n`as_stream: true` is passed in the `opts` argument.","ref":"Cinema.Projection.Lens.html#apply/3","title":"Cinema.Projection.Lens.apply/3","type":"function"},{"doc":"Builds a new `Elixir.Cinema.Projection.Lens.t()` struct with the given `filters` and optional `reducer`.\nSee `Elixir.Cinema.Projection.Lens.apply/3` for more information.","ref":"Cinema.Projection.Lens.html#build!/2","title":"Cinema.Projection.Lens.build!/2","type":"function"},{"doc":"","ref":"Cinema.Projection.Lens.html#t:t/0","title":"Cinema.Projection.Lens.t/0","type":"type"},{"doc":"This module is responsible for ranking a list of projections in a depth-first manner, which can\nbe used to determine the order in which to execute a list of projections, potentially asynchronously.\n\nThe `derive/2` function is the main entry point, which takes a list of projections and a single\ntarget projection you'd like to run. The function will then proceed to rank all dependencts of the\ntarget projection recursively.\n\nNote that a single dependency can appear multiple times due to being dependents of different subgraphs\nof the target projection. In this case, the dependency will be ranked multiple times, but the final\nranking will be the highest rank of all duplicate dependencies.\n\nRanked projections can then be grouped by rank, and the groups can be executed in parallel, as they\nare guaranteed to be independent of each other.\n\nSee the `depth_first_rank/2` function for more details on how the ranking is actually done.","ref":"Cinema.Projection.Rank.html","title":"Cinema.Projection.Rank","type":"module"},{"doc":"Recursively traverses a list of KV tuples representing a graph of projections and their dependencies,\nand ranks them in a depth-first manner.\n\nDoes not do any additional processing on the ranked projections, see `derive/2` for that.","ref":"Cinema.Projection.Rank.html#depth_first_rank/2","title":"Cinema.Projection.Rank.depth_first_rank/2","type":"function"},{"doc":"Derives the dependency order of a target projection in a list of projections by ranking them in a\ndepth-first manner.\n\nProjections are expected to be given in the form: `[{Module.t(), [Module.t()]}]`\n\nNote that the order of the returned list is derived simply from the declared hierarchy of the projections,\nand does not take into account any additional processing that may be required to actually execute\nthe projections. This means that the order of the returned list may not be the optimal order for\nexecution, but it will guarantee that all dependencies are executed before the target projection.\n\nAdditionally, the algorithm used is simple in that it \"lazily\" ranks dependencies after the whole graph\nhas been traversed, and does not attempt to optimize the ranking in any way. A graph could theoretically\ntake into consideration \"re-ranking\" of dependencies in an attempt to optimize ordering/grouping, but\nthis is not currently implemented.\n\nAlso do note that currently this function does not handle cycles in the graph, and will raise an error\nif a cycle is detected.\n\nSee examples:\n\n  ```elixir\n  iex> alias Elixir.Cinema.Projection.Rank\n  iex> graph = [a: [], b: [], c: [:a, :b], d: [:c], e: [:b], f: [:c, :e]]\n  iex> Rank.derive!(graph, :d)\n  [[:a, :b], :c, :d]\n  iex> Rank.derive!(graph, :e)\n  [:b, :e]\n  iex> Rank.derive!(graph, :a)\n  [:a]\n  iex> # This isn't neccessarily the optimal order, but it is _correct_.\n  iex> Rank.derive!(graph, :f)\n  [[:a, :b], [:c, :e], :f]\n  ```","ref":"Cinema.Projection.Rank.html#derive!/2","title":"Cinema.Projection.Rank.derive!/2","type":"function"}],"producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}